==> jmode-0.4.20.tar.gz: jmode-0.4.20/src/ximim.cpp <==
// XIM ץȥIMνԤ

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <map>
#include "xim.h"
#include <X11/Xutil.h>
#define NEED_EVENTS	// xEventΤᡢ
#include <X11/Xproto.h>

#ifndef __GNUC__
# ifdef HAVE_ALLOCA_H
#  include <alloca.h>
# endif
#endif

static map<int,ximim *> g_ims;

// emacsΥХ褱
class ForwardingStat{
public:
    ForwardingStat();
    void event(int );
    int repeat;
    int last_time;
    static bool GuiltyClient;
    static char *Message;
};

bool ForwardingStat::GuiltyClient=false;
char *ForwardingStat::Message =
"(¿ʬ)emacs ʥѥåȤƤޤ\n"
"emacsܸʳLANGǵư뤫\n"
"⤷ϥߥʥǵưƤ\n"
"(뤤XIMбѥå)";

class ximim_impl : public ximim{
public:
    ximim_impl(Connection *c,int id);
    virtual ~ximim_impl();
    void create_ic(RxPacket *);
    void destroy_ic(int );
    void set_ic_focus(int icid);
    void set_ic_values(RxPacket *);
    void get_ic_values(RxPacket *);
    void unset_ic_focus(int icid);
    void forward_event(RxPacket *);
    void send_sync_reply(int icid);
    ximic *get_ic_by_id(int id);
    void onSendPacket();

private:
    int unused_ic_id();
    void free_all_ic();
    void delete_ic(ximic *);
    bool check_event(int t);


    map<int,ximic *> m_ics;
    ForwardingStat mFs;
};

ForwardingStat::ForwardingStat()
{
    last_time = 0;
    repeat = 0;
}

void ForwardingStat::event(int t)
{
    if ( last_time == t ){
	repeat ++;
    }else{
	last_time = t;
	repeat = 0;
    }
}

ximim_impl::ximim_impl(Connection *c,int id) :ximim(c,id)
{
}

ximim_impl::~ximim_impl()
{
    free_all_ic();
}

void ximim_impl::create_ic(RxPacket *p)
{
    ximic *ic;
    int icid= unused_ic_id();
    ic = ::create_ic(m_conn,p,m_id,icid);
    if ( !ic ){
	m_conn->push_error_packet(m_id,icid,
				  ERR_Style,"invalid im style");
	m_conn->terminate();
	return ;
    }
    pair<int,ximic *> n(ic->get_icid(),ic);
    m_ics.insert(n);

    TxPacket *t;
    t = createTxPacket(XIM_CREATE_IC_REPLY,0);
    t->pushC16(m_id);
    t->pushC16(ic->get_icid());
    m_conn->push_packet(t);
}

void ximim_impl::destroy_ic(int icid)
{
    TxPacket *t;
    t = createTxPacket(XIM_DESTROY_IC_REPLY,0);
    t->pushC16(m_id);
    t->pushC16(icid);
    m_conn->push_packet(t);
    //IC˲
    ximic *ic;
    ic = get_ic_by_id(icid);
    delete_ic(ic);
}

void ximim_impl::set_ic_values(RxPacket *p)
{
    int imid,icid;
    ximic *ic;
    p->rewind();
    imid = p->getC16();
    icid = p->getC16();
    ic = get_ic_by_id(icid);

    int atr_len;
    atr_len = p->getC16();
    p->getC16();

    unsigned char *v;
    v = (unsigned char *)alloca(atr_len);

    int i;
    for ( i = 0 ; i < atr_len ; i++){
	v[i] = p->getC8();
    }

    ic->set_ic_attrs((void *)v,atr_len);

    TxPacket *t=createTxPacket(XIM_SET_IC_VALUES_REPLY,0);
    t->pushC16(imid);
    t->pushC16(icid);
    m_conn->push_packet(t);
}

void ximim_impl::get_ic_values(RxPacket *p)
{
    int icid;
    ximic *ic;
    p->rewind();
    p->getC16();
    icid = p->getC16();
    ic = get_ic_by_id(icid);

    int len;
    len = p->getC16();

    TxPacket *t = createTxPacket(XIM_GET_IC_VALUES_REPLY,0);
    t->pushC16(m_id);
    t->pushC16(icid);
    int i,id,l;
    l = 0;
    for (i = 0; i < len /2 ; i++){
	id = p->getC16();
	l += ic->get_ic_atr(id,0);
    }
    t->pushC16(l);
    t->pushC16(0);

    p->rewind();
    p->getC16();//imid
    p->getC16();//icid
    p->getC16();// n

    for (i = 0; i < len /2 ; i++){
	id = p->getC16();
	l += ic->get_ic_atr(id,t);
    }
    m_conn->push_packet(t);

}

int ximim_impl::unused_ic_id()
{
    map<int,ximic *>::iterator i;
    int max_id=1; // input-context ID1Ȥ?
    for ( i = m_ics.begin() ; i != m_ics.end() ; i++){
	if ( max_id <= (*i).first ){
	    max_id = (*i).first+1;
	}
    }
    return max_id;
}

void ximim_impl::set_ic_focus(int icid)
{
    ximic *ic = get_ic_by_id(icid);
    if ( ic ){
	ic->set_focus();
    }
}

void ximim_impl::unset_ic_focus(int icid)
{
    ximic *ic = get_ic_by_id(icid);
    if ( ic ){
	ic->unset_focus();
    }
}

ximic *ximim_impl::get_ic_by_id(int icid)
{
    map<int,ximic *>::iterator i;
    i = m_ics.find(icid);
    if ( i == m_ics.end()){
	return 0;
    }
    return (*i).second;
}

void ximim_impl::forward_event(RxPacket *p)
{
    unsigned char *c;
    int i;
    keyEventX k;
    xEvent ev_raw;

    int imid,icid,flag;

    ximic *ic;
    imid = p->getC16();
    icid = p->getC16();
    flag = p->getC16();
    ic = get_ic_by_id(icid);
    k.serial = p->getC16();

    //xEventΥԡ롣
    c = (unsigned char *)&ev_raw;
    for ( i = 0 ; i < (int)sizeof(ev_raw) ; i++){
	*c = p->getC8();
	c++;
    }
    
    k.ev.type = ev_raw.u.u.type &0x7f;
    k.ev.xany.serial = (k.serial << 16)| ev_raw.u.u.sequenceNumber;
    k.ev.xany.display = g_dpy;
    k.ev.xany.send_event = ev_raw.u.u.type>127;
    

    switch(k.ev.type ){
    case KeyPress:
	//case KeyRelease:
	k.ev.xkey.window =
	    m_conn->to_hl(ev_raw.u.keyButtonPointer.event);
	k.ev.xkey.root =
	    m_conn->to_hl(ev_raw.u.keyButtonPointer.root);
	k.ev.xkey.subwindow =
	    m_conn->to_hl(ev_raw.u.keyButtonPointer.child);
	k.ev.xkey.time =
	    m_conn->to_hs(ev_raw.u.keyButtonPointer.time);
	k.ev.xkey.x =
	    m_conn->to_hl(ev_raw.u.keyButtonPointer.eventX);
	k.ev.xkey.y =
	    m_conn->to_hs(ev_raw.u.keyButtonPointer.eventY);
	k.ev.xkey.x_root =
	    m_conn->to_hs(ev_raw.u.keyButtonPointer.rootX);
	k.ev.xkey.y_root =
	    m_conn->to_hs(ev_raw.u.keyButtonPointer.rootY);
	k.ev.xkey.state = ev_raw.u.keyButtonPointer.state;
	k.ev.xkey.keycode = ev_raw.u.u.detail;
	k.ev.xkey.same_screen = ev_raw.u.keyButtonPointer.sameScreen;
	char  buf[10];
	KeySym ks;
	XLookupString(&k.ev.xkey,buf,10,&ks,0);
	k.state = m_conn->to_hs(ev_raw.u.keyButtonPointer.state);
	k.press = (k.ev.type == KeyPress);
	k.key_sym = ks;
	if ( check_event(k.ev.xkey.time)){
	    ic->OnKeyEvent(k);
	}
	send_sync_reply(icid);
	break;
    default:
	printf("unknown type of forwarded event.(%d)\n",k.ev.type);
	break;
    }
}

void ximim_impl::free_all_ic()
{
    map<int,ximic *>::iterator i;
    for ( i = m_ics.begin() ; i!= m_ics.end() ; i++){
	(*i).second->unset_focus();
	delete (*i).second;
    }
    m_ics.erase(m_ics.begin(),m_ics.end());
}

void ximim_impl::delete_ic(ximic *ic)
{
    map<int,ximic *>::iterator i;
    for ( i = m_ics.begin() ; i!= m_ics.end() ; i++){
	if ( (*i).second == ic ){
	    (*i).second->unset_focus();
	    delete (*i).second;
	    m_ics.erase(i);
	    return ;
	}
    }
}

void ximim_impl::send_sync_reply(int icid)
{
    TxPacket *t = createTxPacket(XIM_SYNC_REPLY,0);
    t->pushC16(m_id);
    t->pushC16(icid);
    m_conn->push_packet(t);
}

void ximim_impl::onSendPacket()
{
    map<int,ximic *>::iterator i;
    for ( i = m_ics.begin() ; i!= m_ics.end() ; i++){
	(*i).second->onSendPacket();
    }
}

bool ximim_impl::check_event(int t)
{
    if ( mFs.repeat == 10 ){
	if ( !mFs.GuiltyClient ){
	    mFs.GuiltyClient = true;
	    show_issue(mFs.Message);
	}
	return false;
    }
    mFs.event(t);
    return true;
}

ximim::ximim(Connection *c,int id)
{
    m_conn = c;
    m_id = id;
}

int unused_im_id()
{
    int max_id;
    map<int,ximim *>::iterator i;
    max_id = 1;
    for ( i = g_ims.begin() ; i != g_ims.end() ; i++){
	if ( (*i).first == max_id ){
	    max_id = (*i).first +1;
	}
    }
    return max_id;
}

ximim *create_im(Connection *c,int id)
{
    ximim *im;
    im = new ximim_impl(c,id);
    pair<int , ximim *> p(id,im);
    g_ims.insert(p);
    return im;
}

ximim *get_im_by_id(int id)
{
    map<int,ximim *>::iterator i;
    i = g_ims.find(id);
    if ( i == g_ims.end()){
	return NULL;
    }
    return (*i).second;
}

void close_im(int id)
{
    ximim *im;

    im = get_im_by_id(id);
    delete im;

    map<int,ximim *>::iterator i;
    i = g_ims.find(id);
    g_ims.erase(i);
}
/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 */

==> jmode-0.4.24.tar.gz: jmode-0.4.24/src/ximim.cpp <==
// XIM ץȥIMνԤ

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <map>
#include "xim.h"
#include <X11/Xutil.h>
#define NEED_EVENTS	// xEventΤᡢ
#include <X11/Xproto.h>

#ifndef __GNUC__
# ifdef HAVE_ALLOCA_H
#  include <alloca.h>
# endif
#endif

static map<int,XimIM *> g_ims;

// emacsΥХ褱
class ForwardingStat{
public:
    ForwardingStat();
    void event(int );
    int repeat;
    int last_time;
    static bool GuiltyClient;
    static char *Message;
};

bool ForwardingStat::GuiltyClient=false;
char *ForwardingStat::Message =
"(¿ʬ)emacs ʥѥåȤƤޤ\n"
"emacsܸʳLANGǵư뤫\n"
"⤷ϥߥʥǵưƤ\n"
"(뤤XIMбѥå)";

class XimIM_impl : public XimIM{
public:
    XimIM_impl(Connection *c,int id);
    virtual ~XimIM_impl();
    void create_ic(RxPacket *);
    void destroy_ic(int );
    void set_ic_focus(int icid);
    void set_ic_values(RxPacket *);
    void get_ic_values(RxPacket *);
    void unset_ic_focus(int icid);
    void forward_event(RxPacket *);
    void send_sync_reply(int icid);
    XimIC *get_ic_by_id(int id);
    void onSendPacket();

private:
    int unused_ic_id();
    void free_all_ic();
    void delete_ic(XimIC *);
    bool check_event(int t);


    map<int,XimIC *> m_ics;
    ForwardingStat mFs;
};

ForwardingStat::ForwardingStat()
{
    last_time = 0;
    repeat = 0;
}

void ForwardingStat::event(int t)
{
    if ( last_time == t ){
	repeat ++;
    }else{
	last_time = t;
	repeat = 0;
    }
}

XimIM_impl::XimIM_impl(Connection *c,int id) :XimIM(c,id)
{
}

XimIM_impl::~XimIM_impl()
{
    free_all_ic();
}

void XimIM_impl::create_ic(RxPacket *p)
{
    XimIC *ic;
    int icid= unused_ic_id();
    ic = ::create_ic(mConn,p,mID,icid);
    if ( !ic ){
	mConn->push_error_packet(mID,icid,
				  ERR_Style,"invalid im style");
	mConn->terminate();
	return ;
    }
    pair<int,XimIC *> n(ic->get_icid(),ic);
    m_ics.insert(n);

    TxPacket *t;
    t = createTxPacket(XIM_CREATE_IC_REPLY,0);
    t->pushC16(mID);
    t->pushC16(ic->get_icid());
    mConn->push_packet(t);
}

void XimIM_impl::destroy_ic(int icid)
{
    TxPacket *t;
    t = createTxPacket(XIM_DESTROY_IC_REPLY,0);
    t->pushC16(mID);
    t->pushC16(icid);
    mConn->push_packet(t);
    //IC˲
    XimIC *ic;
    ic = get_ic_by_id(icid);
    delete_ic(ic);
}

void XimIM_impl::set_ic_values(RxPacket *p)
{
    int imid,icid;
    XimIC *ic;
    p->rewind();
    imid = p->getC16();
    icid = p->getC16();
    ic = get_ic_by_id(icid);

    int atr_len;
    atr_len = p->getC16();
    p->getC16();

    unsigned char *v;
    v = (unsigned char *)alloca(atr_len);

    int i;
    for ( i = 0 ; i < atr_len ; i++){
	v[i] = p->getC8();
    }

    ic->set_ic_attrs((void *)v,atr_len);

    TxPacket *t=createTxPacket(XIM_SET_IC_VALUES_REPLY,0);
    t->pushC16(imid);
    t->pushC16(icid);
    mConn->push_packet(t);
}

void XimIM_impl::get_ic_values(RxPacket *p)
{
    int icid;
    XimIC *ic;
    p->rewind();
    p->getC16();
    icid = p->getC16();
    ic = get_ic_by_id(icid);

    int len;
    len = p->getC16();

    TxPacket *t = createTxPacket(XIM_GET_IC_VALUES_REPLY,0);
    t->pushC16(mID);
    t->pushC16(icid);
    int i,id,l;
    l = 0;
    for (i = 0; i < len /2 ; i++){
	id = p->getC16();
	l += ic->get_ic_atr(id,0);
    }
    t->pushC16(l);
    t->pushC16(0);

    p->rewind();
    p->getC16();//imid
    p->getC16();//icid
    p->getC16();// n

    for (i = 0; i < len /2 ; i++){
	id = p->getC16();
	l += ic->get_ic_atr(id,t);
    }
    mConn->push_packet(t);
}

int XimIM_impl::unused_ic_id()
{
    map<int,XimIC *>::iterator i;
    int max_id=1; // input-context ID1Ȥ?
    for ( i = m_ics.begin() ; i != m_ics.end() ; i++){
	if ( max_id <= (*i).first ){
	    max_id = (*i).first+1;
	}
    }
    return max_id;
}

void XimIM_impl::set_ic_focus(int icid)
{
    XimIC *ic = get_ic_by_id(icid);
    if ( ic ){
	ic->set_focus();
    }
}

void XimIM_impl::unset_ic_focus(int icid)
{
    XimIC *ic = get_ic_by_id(icid);
    if ( ic ){
	ic->unset_focus();
    }
}

XimIC *XimIM_impl::get_ic_by_id(int icid)
{
    map<int,XimIC *>::iterator i;
    i = m_ics.find(icid);
    if ( i == m_ics.end()){
	return 0;
    }
    return (*i).second;
}

void XimIM_impl::forward_event(RxPacket *p)
{
    unsigned char *c;
    int i;
    keyEventX k;
    xEvent ev_raw;

    int imid,icid,flag;

    XimIC *ic;
    imid = p->getC16();
    icid = p->getC16();
    flag = p->getC16();
    ic = get_ic_by_id(icid);
    k.serial = p->getC16();

    //xEventΥԡ롣
    c = (unsigned char *)&ev_raw;
    for ( i = 0 ; i < (int)sizeof(ev_raw) ; i++){
	*c = p->getC8();
	c++;
    }
    
    k.ev.type = ev_raw.u.u.type &0x7f;
    k.ev.xany.serial = (k.serial << 16)| ev_raw.u.u.sequenceNumber;
    k.ev.xany.display = gDpy;
    k.ev.xany.send_event = ev_raw.u.u.type>127;
    

    switch(k.ev.type ){
    case KeyPress:
	//case KeyRelease:
	k.ev.xkey.window =
	    mConn->to_hl(ev_raw.u.keyButtonPointer.event);
	k.ev.xkey.root =
	    mConn->to_hl(ev_raw.u.keyButtonPointer.root);
	k.ev.xkey.subwindow =
	    mConn->to_hl(ev_raw.u.keyButtonPointer.child);
	k.ev.xkey.time =
	    mConn->to_hs(ev_raw.u.keyButtonPointer.time);
	k.ev.xkey.x =
	    mConn->to_hl(ev_raw.u.keyButtonPointer.eventX);
	k.ev.xkey.y =
	    mConn->to_hs(ev_raw.u.keyButtonPointer.eventY);
	k.ev.xkey.x_root =
	    mConn->to_hs(ev_raw.u.keyButtonPointer.rootX);
	k.ev.xkey.y_root =
	    mConn->to_hs(ev_raw.u.keyButtonPointer.rootY);
	k.ev.xkey.state = ev_raw.u.keyButtonPointer.state;
	k.ev.xkey.keycode = ev_raw.u.u.detail;
	k.ev.xkey.same_screen = ev_raw.u.keyButtonPointer.sameScreen;
	char  buf[10];
	KeySym ks;
	XLookupString(&k.ev.xkey,buf,10,&ks,0);
	k.state = mConn->to_hs(ev_raw.u.keyButtonPointer.state);
	k.press = (k.ev.type == KeyPress);
	k.key_sym = ks;
	if ( check_event(k.ev.xkey.time)){
	    ic->OnKeyEvent(k);
	}
	send_sync_reply(icid);
	break;
    default:
	printf("unknown type of forwarded event.(%d)\n",k.ev.type);
	break;
    }
}

void XimIM_impl::free_all_ic()
{
    map<int,XimIC *>::iterator i;
    for ( i = m_ics.begin() ; i!= m_ics.end() ; i++){
	(*i).second->unset_focus();
	delete (*i).second;
    }
    m_ics.erase(m_ics.begin(),m_ics.end());
}

void XimIM_impl::delete_ic(XimIC *ic)
{
    map<int,XimIC *>::iterator i;
    for ( i = m_ics.begin() ; i!= m_ics.end() ; i++){
	if ( (*i).second == ic ){
	    (*i).second->unset_focus();
	    delete (*i).second;
	    m_ics.erase(i);
	    return ;
	}
    }
}

void XimIM_impl::send_sync_reply(int icid)
{
    TxPacket *t = createTxPacket(XIM_SYNC_REPLY,0);
    t->pushC16(mID);
    t->pushC16(icid);
    mConn->push_packet(t);
}

void XimIM_impl::onSendPacket()
{
    map<int,XimIC *>::iterator i;
    for ( i = m_ics.begin() ; i!= m_ics.end() ; i++){
	(*i).second->onSendPacket();
    }
}

bool XimIM_impl::check_event(int t)
{
    if ( mFs.repeat == 10 ){
	if ( !mFs.GuiltyClient ){
	    mFs.GuiltyClient = true;
	    show_issue(mFs.Message);
	}
	return false;
    }
    mFs.event(t);
    return true;
}

XimIM::XimIM(Connection *c,int id)
{
    mConn = c;
    mID = id;
}

int unused_im_id()
{
    int max_id;
    map<int,XimIM *>::iterator i;
    max_id = 1;
    for ( i = g_ims.begin() ; i != g_ims.end() ; i++){
	if ( (*i).first == max_id ){
	    max_id = (*i).first +1;
	}
    }
    return max_id;
}

XimIM *create_im(Connection *c,int id)
{
    XimIM *im;
    im = new XimIM_impl(c,id);
    pair<int , XimIM *> p(id,im);
    g_ims.insert(p);
    return im;
}

XimIM *get_im_by_id(int id)
{
    map<int,XimIM *>::iterator i;
    i = g_ims.find(id);
    if ( i == g_ims.end()){
	return NULL;
    }
    return (*i).second;
}

void close_im(int id)
{
    XimIM *im;

    im = get_im_by_id(id);
    delete im;

    map<int,XimIM *>::iterator i;
    i = g_ims.find(id);
    g_ims.erase(i);
}
/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 */
